#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <rpc/rpc.h>

#include "mount.h"
#include "nfs.h"
#include "libnfs.h"

struct nfsFILE {
	nfs_fh3 handle;
	size_t offset;
};

struct nfsDIR {
	nfs_fh3 handle;
	cookie3 cookie;
	cookieverf3 cookieverf;
	entry3 *entry;
	bool_t eof;
	READDIR3res res;
};

static nfs_fh3 *getHandle(const char *);
static nfs_fh3 *getDirHandle(const char *);
static void copyHandle(nfs_fh3 *, const nfs_fh3 *);

static nfs_fh3 baseHandle;
static nfs_fh3 curHandle;
static CLIENT *nfsClient=NULL;
static CLIENT *mountClient=NULL;

int nfsMount(const char *host, const char *path) {
	mountres3 res;

	if(mountClient!=NULL) {
		if(mountClient->cl_auth!=NULL) {
			auth_destroy(mountClient->cl_auth);
			mountClient->cl_auth=NULL;
		}
		clnt_destroy(mountClient);
	}
	mountClient=clnt_create((char *)host, MOUNT_PROGRAM, MOUNT_V3, "udp");
	if(mountClient==NULL) {
		clnt_pcreateerror(host);
		return -1;
	}

	memset(&res, '\0', sizeof(mountres3));
	if(mountproc3_mnt_3((char *)path, &res, mountClient)!=RPC_SUCCESS) {
		clnt_perror(mountClient, "mount");
		return -1;
	} else if(res.fhs_status!=MNT3_OK)  {
		return -1;
	}
	if(nfsClient!=NULL) {
		if(nfsClient->cl_auth!=NULL) {
			auth_destroy(nfsClient->cl_auth);
			nfsClient->cl_auth=NULL;
		}
		clnt_destroy(nfsClient);
	}
	
	nfsClient=clnt_create((char *)host, NFS_PROGRAM, NFS_V3, "udp");
	if(nfsClient==NULL) {
		clnt_pcreateerror(host);
		return -1;
	}
	copyHandle(&baseHandle, (nfs_fh3 *)&res.mountres3_u.mountinfo.fhandle);
	copyHandle(&curHandle, &baseHandle);

	if(nfsClient->cl_auth!=NULL) {
		auth_destroy(nfsClient->cl_auth);
	}
	nfsClient->cl_auth=authunix_create_default();
	
	return 0;
}

int nfsUmount(const char *path) {
	if(mountClient!=NULL && mountproc3_umnt_3((char *)path, NULL, mountClient)!=RPC_SUCCESS) {
		clnt_perror(mountClient, "umount");
		return -1;
	}
	if(nfsClient!=NULL) {
		if(nfsClient->cl_auth!=NULL) {
			auth_destroy(nfsClient->cl_auth);
			nfsClient->cl_auth=NULL;
		}
		clnt_destroy(nfsClient);
		nfsClient=NULL;
	}
	if(mountClient!=NULL) {
		if(mountClient->cl_auth!=NULL) {
			auth_destroy(mountClient->cl_auth);
			mountClient->cl_auth=NULL;
		}
		clnt_destroy(mountClient);
		mountClient=NULL;
	}
	return 0;
}

int nfsSetuid(uid_t uid) {
	int ret=-1;
	if(nfsClient!=NULL) {
		if(nfsClient->cl_auth!=NULL) {
			auth_destroy(nfsClient->cl_auth);
		}
		nfsClient->cl_auth=authunix_create("kilroy", uid, uid, 0, NULL);
		ret=0;
	}
	return ret;
}

ssize_t nfsRead(nfsFILE *fd, void *buf, size_t count) {
	READ3args args;
	READ3res res;
	ssize_t bytes;

	if(nfsClient==NULL) {
		return -1;
	}
	
	args.file=fd->handle;
	args.offset=fd->offset;
	args.count=count;
	
	memset(&res, 0, sizeof(res));
	if(nfsproc3_read_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "read");
		bytes=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Read failed: %d\n", res.status);
		bytes=-1;
	} else {
		if(res.READ3res_u.resok.eof) {
			/* something should probably be here */
		}
		bytes=res.READ3res_u.resok.count;
		fd->offset+=bytes;
		memcpy(buf, res.READ3res_u.resok.data.data_val, bytes);
	}
	xdr_free((xdrproc_t)xdr_READ3res, (char *)&res);
	
	return bytes;
}

ssize_t nfsWrite(nfsFILE *fd, const void *buf, size_t count) {
	WRITE3args args;
	WRITE3res res;
	ssize_t bytes;
	
	if(nfsClient==NULL) {
		return -1;
	}
	
	args.file=fd->handle;
	args.stable=FILE_SYNC;
	args.offset=fd->offset;
		
	memset(&res, 0, sizeof(res));
	args.data.data_val=(void *)buf;
	args.count=args.data.data_len=count;
			
	if(nfsproc3_write_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "write");
		bytes=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Write failed: %d\n", res.status);
		bytes=-1;
	} else {
		bytes=res.WRITE3res_u.resok.count;
		fd->offset+=bytes;
	}
	xdr_free((xdrproc_t)xdr_WRITE3res, (char *)&res);

	return bytes;
}

int nfsClose(nfsFILE *fd) {
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)&fd->handle);
	free(fd);
	return 0;
}

nfsFILE *nfsOpen(const char *name, int flags) {
	nfsFILE *fp=NULL;

	nfs_fh3 *handle=getHandle(name);
	if(handle==NULL) {
		return NULL;
	}
	
	fp=malloc(sizeof(nfsFILE));
	if(fp==NULL) {
		return NULL;
	}

	fp->offset=0;
	copyHandle(&fp->handle, handle);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);

	return fp;
}

static nfs_fh3 *getHandle(const char *name) {
	nfs_fh3 *handle=NULL;
	char *file;
	char *path;
	char *init;

	if(nfsClient==NULL) {
		return NULL;
	}

	init=path=malloc(strlen(name)+1);
	if(init==NULL) {
		return NULL;
	}
	memcpy(init, name, strlen(name)+1);

	handle=malloc(sizeof(nfs_fh3));
	if(handle==NULL) {
		free(init);
		return NULL;
	}
	if(path[0]=='/') {
		path++;
		copyHandle(handle, &baseHandle);
	} else {
		copyHandle(handle, &curHandle);
	}
	
	while((file=strtok_r(NULL, "/", &path))!=NULL) {
		LOOKUP3args args;
		LOOKUP3res res;
		memset(&res, 0, sizeof(res));
		copyHandle(&args.what.dir, handle);
		xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);
		args.what.name=file;

		if(nfsproc3_lookup_3(args, &res, nfsClient)!=RPC_SUCCESS) {
			clnt_perror(nfsClient, "lookup");
			free(init);
			free(handle);
			xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)&args.what.dir);
			return NULL;
		} else if(res.status!=NFS3_OK) {
			fprintf(stderr, "Lookup failed: %d\n", res.status);
			free(init);
			free(handle);
			xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)&args.what.dir);
			xdr_free((xdrproc_t)xdr_LOOKUP3res, (char *)&res);
			return NULL;
		}
		copyHandle(handle, &res.LOOKUP3res_u.resok.object);
		xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)&args.what.dir);
		xdr_free((xdrproc_t)xdr_LOOKUP3res, (char *)&res);
	}
	
	free(init);
	return handle;
}

static nfs_fh3 *getDirHandle(const char *name) {
	nfs_fh3 *handle=NULL;
	char *file;
	char *path;
	char *init;

	if(nfsClient==NULL) {
		return NULL;
	}

	init=path=malloc(strlen(name)+1);
	if(init==NULL) {
		return NULL;
	}
	memcpy(init, name, strlen(name)+1);

	handle=malloc(sizeof(nfs_fh3));
	if(handle==NULL) {
		free(init);
		return NULL;
	}

	if(path[0]=='/') {
		path++;
		copyHandle(handle, &baseHandle);
	} else {
		copyHandle(handle, &curHandle);
	}
	
	while((file=strtok_r(NULL, "/", &path))!=NULL && path[0]!='\0') {
		LOOKUP3args args;
		LOOKUP3res res;
		memset(&res, 0, sizeof(res));
		copyHandle(&args.what.dir, handle);
		xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);
		args.what.name=file;

		if(nfsproc3_lookup_3(args, &res, nfsClient)!=RPC_SUCCESS) {
			clnt_perror(nfsClient, "lookup");
			free(init);
			free(handle);
			xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)&args.what.dir);
			return NULL;
		} else if(res.status!=NFS3_OK) {
			fprintf(stderr, "Lookup failed: %d\n", res.status);
			free(init);
			free(handle);
			xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)&args.what.dir);
			xdr_free((xdrproc_t)xdr_LOOKUP3res, (char *)&res);
			return NULL;
		}
		copyHandle(handle, &res.LOOKUP3res_u.resok.object);
		xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)&args.what.dir);
		xdr_free((xdrproc_t)xdr_LOOKUP3res, (char *)&res);
	}
	
	free(init);
	return handle;
}

static void copyHandle(nfs_fh3 *dest, const nfs_fh3 *source) {
	dest->data.data_len=source->data.data_len;
	dest->data.data_val=malloc(dest->data.data_len);
	if(dest->data.data_val==NULL) {
		dest->data.data_len=0;
		return;
	}
	memcpy(dest->data.data_val, source->data.data_val, dest->data.data_len);
}

int nfsCreate(const char *name, mode_t mode) {
	CREATE3args args;
	CREATE3res res;
	int ret=0;
	nfs_fh3 *handle;
	mode_t mask=umask(0);
	umask(mask);

	if(nfsClient==NULL) {
		return -1;
	}
	
	handle=getDirHandle(name);
	if(handle==NULL) {
		return -1;
	}
	args.where.dir=*handle;
	args.where.name=strrchr(name, '/');
	if(args.where.name==NULL) {
		args.where.name=(char *)name;
	}
	args.how.mode=GUARDED;

	args.how.createhow3_u.obj_attributes.mode.set_it=TRUE;
	args.how.createhow3_u.obj_attributes.mode.set_mode3_u.mode=mode & ~mask;
	args.how.createhow3_u.obj_attributes.uid.set_it=FALSE;
	args.how.createhow3_u.obj_attributes.gid.set_it=FALSE;
	args.how.createhow3_u.obj_attributes.size.set_it=FALSE;
	args.how.createhow3_u.obj_attributes.atime.set_it=DONT_CHANGE;
	args.how.createhow3_u.obj_attributes.mtime.set_it=DONT_CHANGE;
	
	memset(&res, 0, sizeof(res));
	
	if(nfsproc3_create_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "create");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Create failed: %d\n", res.status);
		ret=-1;
	}
	xdr_free((xdrproc_t)xdr_CREATE3res, (char *)&res);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);
	return ret;
}

int nfsFchmod(nfsFILE *fp, mode_t mode) {
	SETATTR3args args;
	SETATTR3res res;
	int ret=0;
	
	if(nfsClient==NULL) {
		return -1;
	}
	
	args.object=fp->handle;

	args.new_attributes.mode.set_it=TRUE;
	args.new_attributes.mode.set_mode3_u.mode=mode;
	args.new_attributes.uid.set_it=FALSE;
	args.new_attributes.gid.set_it=FALSE;
	args.new_attributes.size.set_it=FALSE;
	args.new_attributes.atime.set_it=DONT_CHANGE;
	args.new_attributes.mtime.set_it=DONT_CHANGE;
	args.guard.check=FALSE;

	memset(&res, 0, sizeof(res));
	
	if(nfsproc3_setattr_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "setattr");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Setattr failed: %d\n", res.status);
		ret=-1;
	}
	xdr_free((xdrproc_t)xdr_SETATTR3res, (char *)&res);
	return ret;
}

int nfsFchown(nfsFILE *fp, uid_t uid) {
	SETATTR3args args;
	SETATTR3res res;
	int ret=0;
	
	if(nfsClient==NULL) {
		return -1;
	}
	
	args.object=fp->handle;

	args.new_attributes.mode.set_it=FALSE;
	args.new_attributes.uid.set_it=TRUE;
	args.new_attributes.uid.set_uid3_u.uid=uid;
	args.new_attributes.gid.set_it=FALSE;
	args.new_attributes.size.set_it=FALSE;
	args.new_attributes.atime.set_it=DONT_CHANGE;
	args.new_attributes.mtime.set_it=DONT_CHANGE;
	args.guard.check=FALSE;
	
	memset(&res, 0, sizeof(res));
	
	if(nfsproc3_setattr_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "setattr");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Setattr failed: %d\n", res.status);
		ret=-1;
	}
	xdr_free((xdrproc_t)xdr_SETATTR3res, (char *)&res);
	return ret;
}

int nfsFchgrp(nfsFILE *fp, gid_t gid) {
	SETATTR3args args;
	SETATTR3res res;
	int ret=0;
	
	if(nfsClient==NULL) {
		return -1;
	}
	
	args.object=fp->handle;

	args.new_attributes.mode.set_it=FALSE;
	args.new_attributes.uid.set_it=FALSE;
	args.new_attributes.gid.set_it=TRUE;
	args.new_attributes.gid.set_gid3_u.gid=gid;
	args.new_attributes.size.set_it=FALSE;
	args.new_attributes.atime.set_it=DONT_CHANGE;
	args.new_attributes.mtime.set_it=DONT_CHANGE;
	args.guard.check=FALSE;
	
	memset(&res, 0, sizeof(res));
	
	if(nfsproc3_setattr_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "setattr");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Setattr failed: %d\n", res.status);
		ret=-1;
	}
	xdr_free((xdrproc_t)xdr_SETATTR3res, (char *)&res);
	return ret;
}

int nfsChmod(const char *path, mode_t mode) {
	int ret;
	nfsFILE *handle=nfsOpen(path, O_RDWR);
	if(handle==NULL) {
		return -1;
	}
	ret=nfsFchmod(handle, mode);
	nfsClose(handle);
	return ret;
}

int nfsChown(const char *path, uid_t uid) {
	int ret;
	nfsFILE *handle=nfsOpen(path, O_RDWR);
	if(handle==NULL) {
		return -1;
	}
	ret=nfsFchown(handle, uid);
	nfsClose(handle);
	return ret;
}

int nfsChgrp(const char *path, gid_t gid) {
	int ret;
	nfsFILE *handle=nfsOpen(path, O_RDWR);
	if(handle==NULL) {
		return -1;
	}
	ret=nfsFchgrp(handle, gid);
	nfsClose(handle);
	return ret;
}

int nfsStat(const char *path, struct stat *buf) {
	int ret;
	nfsFILE *handle=nfsOpen(path, O_RDWR);
	if(handle==NULL) {
		return -1;
	}
	ret=nfsFstat(handle, buf);
	nfsClose(handle);
	return ret;
}

int nfsFstat(nfsFILE *fp, struct stat *buf) {
	GETATTR3args args;
	GETATTR3res res;
	fattr3 *info;

	if(nfsClient==NULL) {
		return -1;
	}
	
	args.object=fp->handle;
	
	memset(&res, '\0', sizeof(res));

	if(nfsproc3_getattr_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "getattr");
		return -1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Getattr failed: %d\n", res.status);
		return -1;
	}
	info=&res.GETATTR3res_u.resok.obj_attributes;

	buf->st_dev=0;
	buf->st_ino=0;
	buf->st_mode=info->mode;
	buf->st_nlink=info->nlink;
	buf->st_uid=info->uid;
	buf->st_gid=info->gid;
	buf->st_rdev=0;
	buf->st_size=info->size;
	buf->st_blksize=0;
	buf->st_blocks=0;
	buf->st_atime=info->atime.seconds;
	buf->st_mtime=info->mtime.seconds;
	buf->st_ctime=info->ctime.seconds;

	switch(info->type) {
	case NFS3REG:
		buf->st_mode |= 0100000;
		break;
	case NFS3DIR:
		buf->st_mode |= 0040000;
		break;
	case NFS3BLK:
		buf->st_mode |= 0060000;
		break;
	case NFS3CHR:
		buf->st_mode |= 0020000;
		break;
	case NFS3LNK:
		buf->st_mode |= 0120000;
		break;
	case NFS3SOCK:
		buf->st_mode |= 0140000;
		break;
	case NFS3FIFO:
		buf->st_mode |= 0010000;
		break;
	}

	xdr_free((xdrproc_t)xdr_GETATTR3res, (char *)&res);
	return 0;
}

nfsDIR *nfsOpendir(const char *path) {
	nfsDIR *dir;
	nfs_fh3 *handle;
	
	dir=malloc(sizeof(nfsDIR));
	if(dir==NULL) {
		return NULL;
	}

	handle=getHandle(path);
	if(handle==NULL) {
		free(dir);
		return NULL;
	}

	copyHandle(&dir->handle, handle);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);

	dir->cookie=0;
	memset(&dir->cookieverf, '\0', sizeof(cookieverf3));
	dir->entry=NULL;
	dir->eof=FALSE;
	memset(&dir->res, '\0', sizeof(dir->res));
	
	return dir;
}

char *nfsReaddir(nfsDIR *dir) {
	static char name[NAME_MAX];
	
	if(nfsClient==NULL || dir==NULL) {
		return NULL;
	}
	
	if(dir->entry==NULL) {
		READDIR3args args;

		xdr_free((xdrproc_t)xdr_READDIR3res, (char *)&dir->res);
		memset(&dir->res, '\0', sizeof(dir->res));
		
		args.dir=dir->handle;
		args.cookie=dir->cookie;
		memcpy(&args.cookieverf, &dir->cookieverf, sizeof(cookieverf3));
		args.count=1024;

		if(nfsproc3_readdir_3(args, &dir->res, nfsClient)!=RPC_SUCCESS) {
			clnt_perror(nfsClient, "readdir");
			return NULL;
		} else if(dir->res.status!=NFS3_OK) {
			fprintf(stderr, "Readdir failed: %d\n", dir->res.status);
			return NULL;
		}

		memcpy(&dir->cookieverf, &dir->res.READDIR3res_u.resok.cookieverf, sizeof(cookieverf3));
		dir->eof=dir->res.READDIR3res_u.resok.reply.eof;
		dir->entry=dir->res.READDIR3res_u.resok.reply.entries;
	}
	if(dir->eof==TRUE && dir->entry==NULL) {
		return NULL;
	}
	dir->cookie=dir->entry->cookie;
	strncpy(name, dir->entry->name, NAME_MAX-1);
	dir->entry=dir->entry->nextentry;
	name[NAME_MAX]='\0';
	return name;
}

int nfsClosedir(nfsDIR *dir) {
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)&dir->handle);
	xdr_free((xdrproc_t)xdr_READDIR3res, (char *)&dir->res);
	free(dir);
	return 0;
}

int nfsMkdir(const char *name, mode_t mode) {
	MKDIR3args args;
	MKDIR3res res;
	int ret=0;
	nfs_fh3 *handle;
	mode_t mask=umask(0);
	umask(mask);
	
	if(nfsClient==NULL) {
		return -1;
	}
	
	handle=getDirHandle(name);
	if(handle==NULL) {
		return -1;
	}
	args.where.dir=*handle;
	args.where.name=strrchr(name, '/');
	if(args.where.name==NULL) {
		args.where.name=(char *)name;
	}

	args.attributes.mode.set_it=TRUE;
	args.attributes.mode.set_mode3_u.mode=mode & ~mask;
	args.attributes.uid.set_it=FALSE;
	args.attributes.gid.set_it=FALSE;
	args.attributes.size.set_it=FALSE;
	args.attributes.atime.set_it=DONT_CHANGE;
	args.attributes.mtime.set_it=DONT_CHANGE;
	
	memset(&res, 0, sizeof(res));
	
	if(nfsproc3_mkdir_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "mkdir");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Mkdir failed: %d\n", res.status);
		ret=-1;
	}
	xdr_free((xdrproc_t)xdr_MKDIR3res, (char *)&res);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);
	return ret;
}

int nfsChdir(const char *name) {
	LOOKUP3args args;
	LOOKUP3res res;
	int ret=0;
	nfs_fh3 *handle;

	if(nfsClient==NULL) {
		return -1;
	}
	
	handle=getDirHandle(name);
	if(handle==NULL) {
		return -1;
	}
	args.what.dir=*handle;
	args.what.name=strrchr(name, '/');
	if(args.what.name==NULL) {
		args.what.name=(char *)name;
	}

	memset(&res, 0, sizeof(res));
	
	if(nfsproc3_lookup_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "lookup");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Lookup failed: %d\n", res.status);
		ret=-1;
	} else {
		copyHandle(&curHandle, &res.LOOKUP3res_u.resok.object);
	}
	xdr_free((xdrproc_t)xdr_LOOKUP3res, (char *)&res);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);
	return ret;
}

int nfsUnlink(const char *name) {
	REMOVE3args args;
	REMOVE3res res;
	int ret=0;
	nfs_fh3 *handle;

	if(nfsClient==NULL) {
		return -1;
	}

	handle=getDirHandle(name);
	if(handle==NULL) {
		return -1;
	}
	args.object.dir=*handle;
	args.object.name=strrchr(name, '/');
	if(args.object.name==NULL) {
		args.object.name=(char *)name;
	}

	memset(&res, 0, sizeof(res));

	if(nfsproc3_remove_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "remove");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Remove failed: %d\n", res.status);
		ret=-1;
	}
	xdr_free((xdrproc_t)xdr_REMOVE3res, (char *)&res);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);
	return ret;
}

int nfsRmdir(const char *name) {
	RMDIR3args args;
	RMDIR3res res;
	int ret=0;
	nfs_fh3 *handle;

	if(nfsClient==NULL) {
		return -1;
	}

	handle=getDirHandle(name);
	if(handle==NULL) {
		return -1;
	}
	args.object.dir=*handle;
	args.object.name=strrchr(name, '/');
	if(args.object.name==NULL) {
		args.object.name=(char *)name;
	}

	memset(&res, 0, sizeof(res));

	if(nfsproc3_rmdir_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "rmdir");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Rmdir failed: %d\n", res.status);
		ret=-1;
	}
	xdr_free((xdrproc_t)xdr_RMDIR3res, (char *)&res);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);
	return ret;
}

int nfsReadlink(const char *path, char *buf, size_t bufsize) {
	READLINK3args args;
	READLINK3res res;
	nfs_fh3 *handle;
	int ret=0;

	if(nfsClient==NULL) {
		return -1;
	}
	
	handle=getHandle(path);
	if(handle==NULL) {
		return -1;
	}
	args.symlink=*handle;
	
	memset(&res, '\0', sizeof(res));
	if(nfsproc3_readlink_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "readlink");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Readlink failed: %d\n", res.status);
		ret=-1;
	} else {
		if(strlen(res.READLINK3res_u.resok.data)>=bufsize) {
			ret=-1;
		} else {
			strcpy(buf, res.READLINK3res_u.resok.data);
			ret=strlen(buf);
		}
	}

	xdr_free((xdrproc_t)xdr_READLINK3res, (char *)&res);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);
	return ret;
}

int nfsRename(const char *src, const char *dest) {
	RENAME3args args;
	RENAME3res res;
	nfs_fh3 *sHandle;
	nfs_fh3 *dHandle;
	int ret=0;

	if(nfsClient==NULL) {
		return -1;
	}
	
	sHandle=getDirHandle(src);
	if(sHandle==NULL) {
		return -1;
	}
	dHandle=getDirHandle(dest);
	if(dHandle==NULL) {
		return -1;
	}

	args.from.dir=*sHandle;
	args.from.name=strrchr(src, '/');
	if(args.from.name==NULL) {
		args.from.name=(char *)src;
	}

	args.to.dir=*dHandle;
	args.to.name=strrchr(src, '/');
	if(args.to.name==NULL) {
		args.to.name=(char *)dest;
	}

	memset(&res, '\0', sizeof(res));

	if(nfsproc3_rename_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "rename");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Rename failed: %d\n", res.status);
		ret=-1;
	}

	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)sHandle);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)dHandle);
	xdr_free((xdrproc_t)xdr_RENAME3res, (char *)&res);
	return ret;
}

int nfsSymlink(const char *src, const char *dest) {
	SYMLINK3args args;
	SYMLINK3res res;
	nfs_fh3 *handle;
	int ret=0;

	if(nfsClient==NULL) {
		return -1;
	}
	
	args.symlink.symlink_data=(char *)src;
	handle=getDirHandle(dest);
	if(handle==NULL) {
		return -1;
	}

	args.where.dir=*handle;
	args.where.name=strrchr(dest, '/');
	if(args.where.name==NULL) {
		args.where.name=(char *)dest;
	}

	args.symlink.symlink_attributes.mode.set_it=TRUE;
	args.symlink.symlink_attributes.mode.set_mode3_u.mode=0777;
	args.symlink.symlink_attributes.uid.set_it=FALSE;
	args.symlink.symlink_attributes.gid.set_it=FALSE;
	args.symlink.symlink_attributes.size.set_it=FALSE;
	args.symlink.symlink_attributes.atime.set_it=DONT_CHANGE;
	args.symlink.symlink_attributes.mtime.set_it=DONT_CHANGE;

	memset(&res, '\0', sizeof(res));
	
	if(nfsproc3_symlink_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "symlink");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Symlink failed: %d\n", res.status);
		ret=-1;
	}

	xdr_free((xdrproc_t)xdr_SYMLINK3res, (char *)&res);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);
	return ret;
}

int nfsLink(const char *src, const char *dest) {
	LINK3args args;
	LINK3res res;
	nfs_fh3 *sHandle;
	nfs_fh3 *dHandle;
	int ret=0;

	if(nfsClient==NULL) {
		return -1;
	}
	
	sHandle=getHandle(src);
	if(sHandle==NULL) {
		return -1;
	}
	args.file=*sHandle;
	dHandle=getDirHandle(dest);

	args.link.dir=*dHandle;
	args.link.name=strrchr(dest, '/');
	if(args.link.name==NULL) {
		args.link.name=(char *)dest;
	}

	memset(&res, '\0', sizeof(res));
	
	if(nfsproc3_link_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "link");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Link failed: %d\n", res.status);
		ret=-1;
	}

	xdr_free((xdrproc_t)xdr_LINK3res, (char *)&res);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)dHandle);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)sHandle);
	return ret;
}

int nfsMknod(const char *name, mode_t mode, dev_t dev) {
	MKNOD3args args;
	MKNOD3res res;
	int ret=0;
	nfs_fh3 *handle;
	mode_t mask=umask(0);
	umask(mask);

	if(nfsClient==NULL) {
		return -1;
	}
	
	handle=getDirHandle(name);
	if(handle==NULL) {
		return -1;
	}
	args.where.dir=*handle;
	args.where.name=strrchr(name, '/');
	if(args.where.name==NULL) {
		args.where.name=(char *)name;
	}

	memset(&res, 0, sizeof(res));
	
	if(nfsproc3_mknod_3(args, &res, nfsClient)!=RPC_SUCCESS) {
		clnt_perror(nfsClient, "mknod");
		ret=-1;
	} else if(res.status!=NFS3_OK) {
		fprintf(stderr, "Mknod failed: %d\n", res.status);
		ret=-1;
	}
	xdr_free((xdrproc_t)xdr_MKNOD3res, (char *)&res);
	xdr_free((xdrproc_t)xdr_nfs_fh3, (char *)handle);
	return ret;
}
